import math
from datetime import datetime
from functools import reduce
import tempfile
import numpy as np
import os
from tqdm import tqdm

import groclient.lib
from api.client.samples.similar_regions import transform

CACHE_PATH = "similar_region_state_cache"
# How much to weight the lowest weight feature.
# The features (coefficients for FFT) per metric will be weighted from 1.0 to LOWEST_PERCENTAGE_WEIGHT_FEATURE
LOWEST_PERCENTAGE_WEIGHT_FEATURE = 0.6

class SimilarRegionState(object):
    """
    Holds and initializes parameters and provide ssaving/loading logic for a similar_region search.
    """

    def __init__(self, region_properties, regions_to_compare, client, data_dir=None, no_download=False):
        self.client = client
        self._logger = groclient.lib.get_default_logger()

        # Figure out temporary directory
        if data_dir:
            self.data_dir = data_dir
        elif os.path.isdir(CACHE_PATH) and os.access(CACHE_PATH, os.W_OK):
            self.data_dir = CACHE_PATH
        else:
            self.data_dir = tempfile.gettempdir()

        self.region_properties = region_properties
        self.num_regions = len(regions_to_compare)
        self.num_properties = len(region_properties)
        self.tot_num_features = reduce(lambda acc, prop: acc + prop["properties"]["num_features"],
                                       region_properties.values(),
                                       0)

        # Mapping to and from our internal numpy idxs
        self.mapping = {region_idx:idx for (idx,region_idx) in enumerate(regions_to_compare)}
        self.inverse_mapping = np.array(regions_to_compare)

        # Data stores and views of this data
        structure = [(name, 'd', p["properties"]["num_features"]) for name, p in
                     region_properties.items()]
        
        # Numpy gives "future deprecation" warning about structured type specs of the form
        # (name,'d',1), i.e. where type is really just a single integer
        # Change such cases into (name,'d')
        structure = [item if item[2] != 1 else (item[0], item[1]) for item in structure]
        
        structure_bool = [(name, bool) for name, p in
                     region_properties.items()]
        self.data = np.ma.zeros(self.num_regions, dtype=structure)
        self.data[:] = np.ma.masked

        self._logger.debug("structure of data array entries is {}".format(structure))
        self._logger.debug("structure of missing array entries is {}".format(structure_bool))

        # Boolean array representing any data we haven't yet downloaded.
        # True if data for that (metric, region) has NOT been fetched into the data array.
        self.missing = np.full(self.num_regions, True, dtype=structure_bool)

        self._create_views()

        # Standardized (0-mean,1-std) version of data in a 2d array (without structure/masking nonsense)
        self.data_standardized = np.zeros((self.num_regions, self.tot_num_features), dtype='d')

        # Weights of each metric
        self.weight_vector = np.ones(self.tot_num_features, dtype='d')
        self._generate_weight_vector()

        self.load(no_download)
        # No modification to the cache if we did not even try to download anything
        if not no_download:
            self.save()

    def _create_views(self):
        self.data_nonstruc = self.data.view(
            dtype=[('data', 'd', (self.num_regions, self.tot_num_features))], type=np.ndarray
        )[0][0]
        self.data_mask_nonstruc = self.data.mask.view(dtype=(bool, self.tot_num_features))

    def save(self):
        """
        Cache the current state to disk.
        This is called AFTER _standardize which means
        all missing data already squeezed out of self.data, self.missing, and self.inverse_mapping
        Safe to simply augment cache with these if not already present
        :return: True if succeeded.
        """
        self._logger.info("Starting caching of downloaded data.")
        
        for name in self.region_properties:
            path = os.path.join(self.data_dir, "{}.nbz".format(name))
            if os.path.isfile(path):
                # we already have a cache => add new regions to it
                with open(path, 'rb') as f:
                    variables = np.load(f)
                    uncached_regions = set(self.inverse_mapping) - set(variables["inverse_mapping"])
                    if uncached_regions:
                        uncached_regions_idx = np.array([self.mapping[idx] for idx in uncached_regions])
                        full_data = np.append(variables["data"], self.data[name][uncached_regions_idx].data, axis=0)
                        full_mask = np.append(variables["mask"], self.data[name][uncached_regions_idx].mask, axis=0)
                        full_missing = np.append(variables["missing"], self.missing[name][uncached_regions_idx], axis=0)
                        full_inverse_mapping = np.append(variables["inverse_mapping"], self.inverse_mapping[uncached_regions_idx], axis=0)
                    else:
                        # no modifications to chache needed
                        continue
            else:
                # just cache entire (newly downloaded) dataset ...
                full_data = self.data[name].data
                full_mask = self.data[name].mask
                full_missing = self.missing[name]
                full_inverse_mapping = self.inverse_mapping
                
            with open(path, 'wb') as f:
                np.savez(f,
                         data=full_data, #self.data[name].data,
                         mask=full_mask, #self.data[name].mask,
                         missing=full_missing, #self.missing[name],
                         inverse_mapping=full_inverse_mapping) #self.inverse_mapping)
        self._logger.info("Done caching of downloaded data.")
        return

    def load(self, no_download):
        """
        Attempt to load any cached information and merge it into our current data situation.
        """
        self._logger.info("Loading data")
        # Loop through the metric views...
        for name in self.region_properties:
            path = os.path.join(self.data_dir, "{}.nbz".format(name))
            assert (not no_download) or os.path.isfile(path), "--no_download requires cached properties to be available" 
            if os.path.isfile(path):
                self._logger.info("Found cached data for {}, loading...".format(name))
                with open(path, 'rb') as f:
                    variables = np.load(f)
                    mutual_regions = set(variables["inverse_mapping"]) & set(self.inverse_mapping)
                    if mutual_regions:
                        # indexes of cached regions of interest in current data 
                        mutual_regions_mapping = np.array([self.mapping[idx] for idx in mutual_regions])
                        old_mapping = {region_idx:idx
                                       for (idx,region_idx) in enumerate(variables["inverse_mapping"])}
                        # indexes of cached regions of interest in cached data
                        mutual_regions_old_mapping = [old_mapping[idx] for idx in mutual_regions]
                        # copy from cached data to new data
                        self.data[name][mutual_regions_mapping] = variables["data"][mutual_regions_old_mapping]
                    
                        mutual_regions_masked = variables["mask"][mutual_regions_old_mapping]
                        self.data[name][mutual_regions_mapping][mutual_regions_masked] = np.ma.masked
                    
                        # cached regions of interest have data (missing=False) UNLESS
                        # they are marked as True in 'missing'
                        self.missing[name][mutual_regions_mapping] = False
                        mutual_regions_missing = variables["missing"][mutual_regions_old_mapping]
                        self.missing[name][mutual_regions_mapping[mutual_regions_missing]] = True
                self._logger.info("Loaded {} cached regions for property {}".format(len(mutual_regions), name))
            # Fill in the missing data if any e.g. in case the cache
            # was created with a subset of current regions_to_compare
            if not no_download:
                self._get_data(name)
        self._standardize()
        self._logger.info("Done loading.")
        return

    def _generate_weight_vector(self):
        # generate weight vector (with decreasing weights if enabled)
        # iterating over the dictionary is fine as we don't allow it
        # to be modified during runtime.  Distance between points is
        # p=2 weighted Minkowski distance, i.e. dist(x,y) = sqrt(\sum
        # [w_i(x_i - y_i)]^2) so sqrt of user-provided weight is taken
        # BEFORE individual property is scaled by its weight as it is
        # assumed user wants dist^2 = \sum w_user_i*(x_i - y_i)^2
        # (note user weight is NOT squared)
        #
        # Alternatively - do not rescale properties, just specify
        # 'wminkowski' metrics for BallTree constructor
        progress_idx = 0
        for properties in self.region_properties.values():
            num_features = properties["properties"]["num_features"]
            # User-provided weight is uniformly spread over
            # num_features (sum will be less for F-transformed
            # properties if weight_slope is requested
            weight_per_feature = (float(properties["properties"]["weight"])/num_features)**0.5
            if properties["properties"]["type"] == "timeseries_fourier" and properties["properties"]["weight_slope"]:
                slope_vector = np.arange(1.0, LOWEST_PERCENTAGE_WEIGHT_FEATURE,
                                         -(1.0 - LOWEST_PERCENTAGE_WEIGHT_FEATURE) / float(num_features))
                slope_vector *= weight_per_feature
                self.weight_vector[progress_idx:progress_idx+num_features] = slope_vector
            else:
                self.weight_vector[progress_idx:progress_idx + num_features] *= weight_per_feature
            progress_idx += num_features

    def _standardize(self):
        self._logger.info("Standardizing data matrix...")

        rows_to_keep = ~np.any(self.data_mask_nonstruc, axis=1)
        # Remove any rows that are missing....
        self.data = self.data[rows_to_keep]
        self.missing = self.missing[rows_to_keep]
        self.data_standardized = self.data_standardized[rows_to_keep]
        # Remove this from the region mapping
        self.inverse_mapping = self.inverse_mapping[rows_to_keep]
        # Recreate the mapping in other direction
        self.mapping.clear()
        for idx, region_idx in enumerate(self.inverse_mapping):
            self.mapping[region_idx] = idx
        # update the number of regions
        self.num_regions = len(self.inverse_mapping)

        # update nonstruc from data array
        self._create_views()

        # copy in from the nonstruc view
        np.copyto(self.data_standardized, self.data_nonstruc)
        mean = np.ma.average(self.data_nonstruc, axis=0)
        std = np.ma.std(self.data_nonstruc, axis=0)

        self._logger.debug("(mean, std) of data across regions axis are ({}, {})".format(mean, std))

        self.data_standardized -= mean
        self.data_standardized /= std
        self.data_standardized *= self.weight_vector

        # TODO: handle missing data properly.
        #https://math.stackexchange.com/questions/195245/average-distance-between-random-points-on-a-line-segment

        self._logger.info("Done standardizing data matrix.")
        return

    def _get_data(self, property_name):
        """Gets data for all regions for the given property and saves it to
        memory (and local cache on disk).
        """
        props = self.region_properties[property_name]
        query = props["selected_entities"]
        # Let's ask the server what times we have available and use those in post-processing.
        series_list = self.client.get_data_series(**query)
        assert len(series_list) > 0, "No data series found for selection {}".format(query)
        data_series = series_list[0]
        if props["properties"]["type"] == "timeseries_fourier":
            start_date = datetime.strptime(data_series["start_date"], '%Y-%m-%dT%H:%M:%S.%fZ')
            period_length_days = self.client.lookup('frequencies', query["frequency_id"])['periodLength']['days']
            end_date = datetime.strptime(data_series["end_date"], '%Y-%m-%dT%H:%M:%S.%fZ')
            num_of_points = (end_date - start_date).days / period_length_days
            self._logger.info("length of data series is {} periodLengths (per frequency_id)".format(num_of_points))
            longest_period_feature_period = props["properties"]["longest_period_feature_period"]
            start_idx = math.floor(num_of_points / float(longest_period_feature_period))
            self._logger.info("first coef index will be {}".format(start_idx))
        num_features = props["properties"]["num_features"]
        # deep copy the metric for each query.
        queries = []
        map_query_to_data_table = []
        for region in self.inverse_mapping[self.missing[property_name]]:
            copy_of_metric = dict(query)
            copy_of_metric["region_id"] = region
            queries.append(copy_of_metric)
            map_query_to_data_table.append(self.mapping[region])

        load_bar = tqdm(total=len(queries))

        def map_response(idx, _, response, *args):
            data_table_idx = map_query_to_data_table[idx]
            if response is None or len(response) == 0 or (len(response) == 1 and response[0] == {}):
                # no data on this region. let's fill it with zeros for the odd cases and mark it for masking.
                self.data[property_name][data_table_idx] = 0.0
                # flag this as invalid.
                self.data[property_name][data_table_idx] = np.ma.masked
            else:
                if props["properties"]["type"] == "timeseries_fourier":
                    # TODO: remove start_datetime stuff here once we have "addNulls" available in api
                    result, coverage = transform.post_process_timeseries(num_of_points, start_date, response,
                                                                         start_idx, num_features,
                                                                         period_length_days=period_length_days)
                    # if there are less points than there are in our lowest period event, let's discard this...
                    if coverage < 1/float(start_idx):
                        self.data[property_name][data_table_idx] = 0.0
                        # flag this as invalid.
                        self.data[property_name][data_table_idx] = np.ma.masked
                    else:
                        self.data[property_name][data_table_idx] = result
                elif props["properties"]["type"] == "pit":
                    # for point in time just add the value
                    if response[0]["value"] is None or np.isnan(response[0]["value"]):
                        self.data[property_name][data_table_idx] = np.ma.masked
                    else:
                        self.data[property_name][data_table_idx] = response[0]["value"]
            # Mark this as downloaded.
            self.missing[property_name][data_table_idx] = False
            load_bar.update()

        self._logger.info("Getting data series for {} regions for property {}".format(len(queries), property_name))
        self.client.batch_async_get_data_points(queries, map_result=map_response)
        load_bar.close()
        return





"""
This library contains utilities to extract features from
time series
"""

import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA
from tsfresh import extract_features


def ts_feature_extraction(dataframe, num_jobs=0):
    """
    Gets 5 transformed features from 794 features extracted by tsfresh
    :param dataframe: A pandas dataframe
    :param num_jobs: integer, number of parallel processes in tsfresh
    :return: A pandas dataframe
    """
    features = extract_features(dataframe,
                                column_id='period',
                                column_sort='date',
                                column_value='value',
                                n_jobs=num_jobs
                                )
    features = rm_const_cols(features)
    return pca_transformation(features)


def rm_const_cols(dataframe):
    """
    Removes dataframe columns with constant values or null any value
    :param dataframe: A pandas dataframe
    :return: A pandas dataframe
    """
    dataframe = dataframe.dropna(axis=1)  # dropping columns with any null value for PCA
    return dataframe.loc[:, (dataframe != dataframe.iloc[0]).any()]


def feature_scaling(dataframe):
    """
    z-transforms the columns
    :param dataframe: A pandas dataframe
    :return: A pandas dataframe
    """
    # Replace infinities by nulls to finally drop them
    dataframe = dataframe.replace([np.inf, -np.inf], np.nan)
    dataframe = rm_const_cols(dataframe)
    scaler = StandardScaler()
    return scaler.fit_transform(dataframe)


def pca_transformation(dataframe, components=5):
    """
    Computes and retrieves the following number of
    principal component columns given by min(5PCs, 80% variation)
    :param dataframe: A pandas dataframe
    :param components: A positive integer
    :return: A pandas dataframe
    """
    if dataframe.isnull().values.any():
        raise ValueError('Null values in the dataframe, cannot perform PCA')
    # PCA components is bounded above by number of samples
    components = min(components, dataframe.shape[0])
    pca = PCA(n_components=components)
    scaled_dataframe = feature_scaling(dataframe)
    features_pca = pd.DataFrame(pca.fit_transform(scaled_dataframe))
    pca_var_cum = np.cumsum(np.round(pca.explained_variance_ratio_, decimals=4) * 100)
    k = components
    for i in range(components - 1, 0, -1):
        if pca_var_cum[i] < 80:
            k = max(i + 1, components)
            break
    df = features_pca.iloc[:, :k]
    df.index = dataframe.index
    return df


def cumulative(dataframe, column_1, column_2):
    """
    Compute cumulative statistics of the 3rd argument after
    grouping by the second argument in a dataframe
    :param dataframe: A pandas dataframe
    :param column_1: String, a column name
    :param column_2: String, a column name
    :return: A pandas dataframe
    """
    return pd.DataFrame(dataframe.groupby([column_1])[column_2].sum())



import numpy as np
import os
from sklearn.neighbors import BallTree
from groclient import GroClient
from groclient.lib import get_default_logger
from api.client.samples.similar_regions.similar_region_state import SimilarRegionState
from sklearn.metrics.pairwise import euclidean_distances

""" API Config """
API_HOST = 'api.gro-intelligence.com'
ACCESS_TOKEN = os.environ['GROAPI_TOKEN']

class SimilarRegion(object):

    def __init__(self, region_properties, regions_to_compare=None, data_dir=None, no_download=False):
        """
        :param region_properties: A dict containing the properties of regions to use when doing the similarity
        computation. This is by default defined in region_properties.py, but can be adjusted.
        :param regions_to_compare: Default is None. This is a list of region IDs against which we want to do the
        similarity computation. If none specified, do this for all regions where we have _any_ data.
        """
        self.client = GroClient(API_HOST, ACCESS_TOKEN)
        self._logger = get_default_logger()
        if not regions_to_compare:
            regions_to_compare = self._regions_avail_for_selection(region_properties)
        self._logger.info("SimilarRegionState: loading...")
        self.state = SimilarRegionState(region_properties, regions_to_compare, self.client, data_dir=data_dir, 
                no_download=no_download)
        self._logger.info("SimilarRegionState: done.")
        self._logger.info("BallTree: computing...")
        # Featues are weighted at this point, so this is equivalent to using
        # sklearn.neighbors.DistanceMetric.get_metric('wminkowski', **{'p':2,'w':self.state.weight_vector})
        # but with unweighted features
        self.ball = BallTree(self.state.data_standardized, leaf_size=2)
        self._logger.info("BallTree: done.")
        return

    def _regions_avail_for_selection(self, region_properties):
        regions = set()
        for props in region_properties.values():
            for available_series in self.client.list_available(props["selected_entities"]):
                regions.add(available_series["region_id"])
        self._logger.info("{} regions are available for comparison.".format(len(regions)))
        return list(regions)

    def _format_results(self, sim_regions, requested_region_level, dists, metric_dists):
        for ranking, sim_region_region_id in enumerate(sim_regions):
            region_info = self.client.lookup("regions", sim_region_region_id)
            region_level = region_info["level"]
            sim_region_name = region_info["name"]
            
            if requested_region_level is not None and region_level != requested_region_level:
                self._logger.info("not level %s %s" % (requested_region_level, sim_region_name))
                continue
                
            # Choose parent one level up from this region
            parent = {"name": ""}
            for r in self.client.lookup_belongs("regions", sim_region_region_id):
                parent = r # in case there is no parent at correct level we will just take the last
                if r['level'] == region_level-1:
                    break

            if parent.get('id', None):
                grandparent = next(self.client.lookup_belongs("regions", parent["id"]), {"name": "", "id": ""})
            else:
                grandparent = {"name": "", "id": ""}
            self._logger.info(u"{}, {}, {}".format(sim_region_name, parent["name"], grandparent["name"]))
            metric_dists_dict = {prop_name: distance for (prop_name, distance) in zip(self.state.region_properties.keys(), metric_dists[ranking])}
            data_point = {"id": sim_region_region_id, "name": sim_region_name, "dist": dists[ranking],
                    "parent": (parent["id"], parent["name"], grandparent["id"], grandparent["name"]), "metric_dist": metric_dists_dict}
            yield data_point

    def _similar_to(self, region_id, number_of_regions):
        """
        This gives us the regions that are most similar to the either the given region in the case on region is give, or
        similar to the "collective mean" of the regions as given.
        :param region_id: a Gro region id representing the reference region you want to find similar regions to.
        :param number_of_regions: number of most similar matches to return
        :return: regions that are most similar to the given region id, the distances to each of these, and the 
                    individual distances for each property (in the order of properties as iterated on the properties
                    dictionary). 
        """
        assert region_id in self.state.mapping, "This region is not available in your configuration or " \
                                                "it lacks coverage in the chosen region properties."
        # list as an index to preserve dimensionality of returned data
        x = self.state.data_standardized[self.state.mapping[region_id], :]
        x = x.reshape(1, -1)
        neighbour_dists, neighbour_idxs = self.ball.query(x, k = min(number_of_regions, self.state.num_regions))

        # get the individual distances 
        metric_distances = [self._get_distances(self.state.mapping[region_id], idx2) for idx2 in neighbour_idxs[0]]

        return self.state.inverse_mapping[neighbour_idxs[0]], neighbour_dists[0], metric_distances

    def _get_distances(self, idx1, idx2):
        """ Returns the distances in each "metric" for the two given rows
            Distances incorporate all aplied weighting
        """
        progress_idx = 0
        distances = []
        for properties in self.state.region_properties.values():
            num_features = properties["properties"]["num_features"]
            subspace_vec1 = self.state.data_standardized[idx1, progress_idx:progress_idx+num_features]
            subspace_vec2 = self.state.data_standardized[idx2, progress_idx:progress_idx+num_features]
            dist = euclidean_distances([subspace_vec1], [subspace_vec2])[0][0]
            distances.append(dist)
            progress_idx += num_features
        return distances

    def similar_to(self, region_id, number_of_regions=10, requested_level=None):
        """
        Attempt to look up the given name and find similar regions.
        :param region_id: a Gro region id representing the reference region you want to find similar regions to.
        :param number_of_regions: number of most similar matches to return
        :return: a generator of the most similar regions as a list in the form
        [{id: 123, name: "abc", distance: 1.23, parent_regions: []]
        """
        sim_regions, dists, metric_dists = self._similar_to(region_id, number_of_regions)
        self._logger.info("Found {} regions most similar to '{}'.".format(len(sim_regions), region_id))
        for output in self._format_results(sim_regions, requested_level, dists, metric_dists):
            yield output



from groclient.lib import get_default_logger

logger = get_default_logger()
logger.warning('''
    Deprecation Warning!

    You are importing modules from deprecated `api` module to access Gro
    Intelligence's API.  Please update your code to import from the `groclient`
    module instead.  The `api` module will be removed by 2021-03-31.

    Replace: from api.client.gro_client import GroClient
       with: from groclient import GroClient

    And replace any other imports from `api.client.*` with imports from
    `groclient.*` instead.

    Please reach out to api-support@gro-intelligence.com if you need any help!
''')



"""Sample API requests for various Gro data series.

This is a very simple collection of Gro API data pulls, designed to
show a sample of the various types of information available and
some syntactical hints for new API users.

The various entity IDs are all aliased to global variable names
in the first section so that the queries themselves read more
intuitively.

After this script runs, the 12 series will be combined in a single data frame
and a preview is printed to the console.

They hold different data as per the comments above each query.

1.International Cocoa Organization (ICCO)
2.International Coffee Organization (ICO)
3.Brazilian Sugarcane Industry Association (UNICA)
4.Center for Advanced Studies on Applied Economics (CEPEA)
5.GIMMS MODIS NDVI
6.MOD11 Land Surface Temperature
7.FEWS NET Daily Evapotranspiration
8.FEWS NET Monthly Actual Evapotranspiration
9.GlobCover Land Cover
10.Soil Moisture and Ocean Salinity Mission (SMOS) Soil Moisture
11.National Farmers Information System (NAFIS) Kenya
12.Gro Yield Model
"""

import os
from groclient import GroClient

# GRO GLOBAL VARIABLES
API_HOST = 'api.gro-intelligence.com'
ACCESS_TOKEN = os.environ['GROAPI_TOKEN']

# METRIC_ID GLOBALS
YIELD = 170037
HARV_AREA = 570001
TEMPERATURE = 2540047
VEG_ANOMALIES = 431132
ETA_PCT = 15852239
PRODUCER_PRICE = 2290065
GRINDINGS = 5380032
COMPOSITE_PRICES = 2270065
ET_VALUE = 4660031
LAND_COVER = 15852316
SOIL_AVAIL = 15531082
WHOLESALE_PX = 2310065

# ITEM_ID GLOBALS
NDVI_DIFF_10YR_MEAN = 321
ETA_PCT_MEDIAN = 4395
LST = 3457
SOIL_MOISTURE = 7382
SOYBEANS = 270
COFFEE = 225
COCOA = 211
COTTON_LINT = 152
SUGARCANE = 538
CASSAVA = 237
BRAZ_ARABICA = 5157
POTENT_ETA = 5072
RAINFED_CROPLAND = 4090
MOISTURE = 7382
AVOCADOS = 127


# REGION_ID GLOBALS
WORLD = 0
BRAZIL = 1029
KENYA_NAIROBI = 13474
UNITED_STATES = 1215
SAO_PAULO = 10408
MATO_GROSSO = 10417

# SOURCE_ID GLOBALS
ICCO = 13
ICO = 9
UNICA = 47
CEPEA = 74
GIMMS_MODIS = 3
LST_SRC = 26
FEWSNET1D = 44
FEWSNET1M = 18
GLOBCOVER = 53
SMOS = 43
NAFIS = 12
GROYM = 32

# FREQUENCY_ID GLOBALS
DAILY = 1
WEEKLY = 2
EIGHTDAY = 3
SIXTEENDAY = 5
MONTHLY = 6
QUARTERLY = 7
ANNUAL = 9
POINT_IN_TIME = 15

# Set up gro API

client = GroClient(API_HOST, ACCESS_TOKEN)

# BRAZIL COCOA GRINDINGS FROM ICCO (1)
client.add_single_data_series({'metric_id': GRINDINGS,
                               'item_id': COCOA,
                               'region_id': BRAZIL,
                               'frequency_id': ANNUAL,
                               'source_id': ICCO})

# BRAZIL ARABICA COMPOSITE PRICES (2)
client.add_single_data_series({'metric_id': COMPOSITE_PRICES,
                               'item_id': BRAZ_ARABICA,
                               'region_id': WORLD,
                               'frequency_id': DAILY,
                               'source_id': ICO})

# SAO PAULO SUGAR AREA FROM UNICA (3)
client.add_single_data_series({'metric_id': HARV_AREA,
                               'item_id': SUGARCANE,
                               'region_id': SAO_PAULO,
                               'frequency_id': ANNUAL,
                               'source_id': UNICA})

# SAO PAULO CASSAVA PRICES FROM CEPEA (4)
client.add_single_data_series({'metric_id': PRODUCER_PRICE,
                               'item_id': CASSAVA,
                               'region_id': SAO_PAULO,
                               'frequency_id': WEEKLY,
                               'source_id': CEPEA})

# NDVI ANOMALY SAO_PAULO (5)
client.add_single_data_series({'metric_id': VEG_ANOMALIES,
                               'item_id': NDVI_DIFF_10YR_MEAN,
                               'region_id': SAO_PAULO,
                               'frequency_id': SIXTEENDAY,
                               'source_id': GIMMS_MODIS})

# LAND SURFACE TEMPERATURE SAO PAULO (6)
client.add_single_data_series({'metric_id': TEMPERATURE,
                               'item_id': LST,
                               'region_id': SAO_PAULO,
                               'frequency_id': DAILY,
                               'source_id': LST_SRC})

# DAILY EVAPOTRANSPIRATION SAO PAULO (7)
client.add_single_data_series({'metric_id': ET_VALUE,
                               'item_id': POTENT_ETA,
                               'region_id': SAO_PAULO,
                               'frequency_id': DAILY,
                               'source_id': FEWSNET1D})

# MONTHLY EVAPOTRANSPIRATION ANOMALY SAO_PAULO (8)
client.add_single_data_series({'metric_id': ETA_PCT,
                               'item_id': ETA_PCT_MEDIAN,
                               'region_id': SAO_PAULO,
                               'frequency_id': MONTHLY,
                               'source_id': FEWSNET1M})

# GLOBCOVER RAINFED CROPLAND SAO_PAULO (9)
client.add_single_data_series({'metric_id': LAND_COVER,
                               'item_id': RAINFED_CROPLAND,
                               'region_id': SAO_PAULO,
                               'frequency_id': POINT_IN_TIME,
                               'source_id': GLOBCOVER})

# SOIL MOISTURE SAO_PAULO (10)
client.add_single_data_series({'metric_id': SOIL_AVAIL,
                               'item_id': MOISTURE,
                               'region_id': SAO_PAULO,
                               'frequency_id': DAILY,
                               'source_id': SMOS})

# NAFIS KENYA AVOCADO PRICE (11)
client.add_single_data_series({'metric_id': WHOLESALE_PX,
                               'item_id': AVOCADOS,
                               'region_id': KENYA_NAIROBI,
                               'frequency_id': DAILY,
                               'source_id': NAFIS})

# GRO US SOYBEAN YIELD (12)
client.add_single_data_series({'metric_id': YIELD,
                               'item_id': SOYBEANS,
                               'region_id': UNITED_STATES,
                               'frequency_id': ANNUAL,
                               'source_id': GROYM})

print(client.get_df())



from __future__ import print_function
# Sample Gro API client for sugar crop model
#
# Usage example:
#
#   export PYTHONPATH=/your/path/to/gro
#   python sugar.py --user_email ...
#
# If you don't want to enter the password each time, print a token,
# save it and use as follows:
#
#   export GROAPI_TOKEN=`python sugar.py --user_email ... --print_token`
#   python sugar.py
#
# Or if you don't save the token, you can pass it via cmd line
#
#   python sugar.py --token ...
#
# Ref: https://app.gro-intelligence.com/#/displays/23713

import argparse
import getpass
import sys
import unicodecsv
import groclient.lib
import os
from api.client.crop_model import CropModel

API_HOST = 'api.gro-intelligence.com'

def main():
    parser = argparse.ArgumentParser(description="Gro api client")
    parser.add_argument("--user_email")
    parser.add_argument("--user_password")
    parser.add_argument("--print_token", action='store_true')
    parser.add_argument("--token", default=os.environ.get('GROAPI_TOKEN', None))
    args = parser.parse_args()
    assert args.user_email or args.token, \
        "Need --token or --user_email"
    access_token = None
    if args.token:
        access_token = args.token
    else:
        if not args.user_password:
            args.user_password = getpass.getpass()
        access_token = groclient.lib.get_access_token(API_HOST, args.user_email, args.user_password)
    if args.print_token:
        print(access_token)
        sys.exit(0)

    model = CropModel(API_HOST, access_token)
    model.add_data_series(item="sugarcane", metric="production quantity", region="Brazil")
    model.add_data_series(item="sugarcane", metric="yield", region="Brazil")
    series_results = model.get_data_series_list()
    for data_series in series_results:
        filename = 'gro_{}_{}_{}.csv'.format(
            model.lookup('items', data_series['item_id']).get('name'),
            model.lookup('metrics', data_series['metric_id']).get('name'),
            model.lookup('regions', data_series['region_id']).get('name'))
        filename = filename.replace('/', ':')
        writer = unicodecsv.writer(open(filename, 'wb'))
        count = 0
        for point in model.get_data_points(**data_series):
            writer.writerow([point['start_date'],
                             point['end_date'],
                             point['value'],
                             model.lookup_unit_abbreviation(point['unit_id'])])
            count += 1
        print("Output {} rows to {}".format(count, filename))

    if series_results:
        data_frame = model.get_df()
        print("Loaded data frame of shape {}, columns: {}".format(data_frame.shape, data_frame.columns))


if __name__ == "__main__":
    main()



import pandas as pd
from pandas.util.testing import assert_frame_equal

from api.client.samples.analogous_years.lib.final_ranks_computation import \
    combined_methods_distances


def data_combined_methods():
    test_data_1 = pd.DataFrame({'rows': ['row_1', 'row_2', 'row_3'],
                                'row_1': [1, 2, 3],
                                'row_2': [4, 5, 6],
                                'row_3': [7, 8, 9]}).set_index('rows')
    test_data_2 = pd.DataFrame({'rows': ['row_1', 'row_2', 'row_3'],
                                'row_4': [1, 2, 3],
                                'row_5': [4, 5, 6],
                                'row_6': [7, 8, 9]}).set_index('rows')
    dictionary_of_df = {'df1': test_data_1, 'df2': test_data_2}
    return dictionary_of_df


def test_combined_methods_distances():
    expected = pd.DataFrame({'rows': ['row_1', 'row_2', 'row_3'],
                             'row_1': [1, 2, 3],
                             'row_2': [4, 5, 6],
                             'row_3': [7, 8, 9],
                             'row_4': [1, 2, 3],
                             'row_5': [4, 5, 6],
                             'row_6': [7, 8, 9],
                             }).set_index('rows')
    assert_frame_equal(combined_methods_distances(data_combined_methods()), expected)



"""Base module for making API requests.

GroClient and CropModel build on top of endpoints exposed in this module.
Helper functions or shims or derivative functionality should appear in the client classes rather
than here.
"""

from builtins import str
from groclient import cfg
from collections import OrderedDict
from groclient.constants import REGION_LEVELS, DATA_SERIES_UNIQUE_TYPES_ID
import groclient.utils
import json
import logging
import requests
import time
import platform
from pkg_resources import get_distribution, DistributionNotFound
try:
    # functools are native in Python 3.2.3+
    from functools import lru_cache as memoize
except ImportError:
    from backports.functools_lru_cache import lru_cache as memoize


# Interpreter and API client library version information.
#
# This is global so we only call get_distribution() once at module load time.
# This is a workaround for a curious bug: in specific situations, calling
# get_distribution() while tornado's event loop is running seems to result in
# GroClient's __del__ method running while the object is still in scope,
# resulting in `fetch() called on closed AsyncHTTPClient` errors upon
# subsequent uses of _async_http_client.
_VERSIONS = { 'python-version': platform.python_version() }
try:
    _VERSIONS['api-client-version'] = get_distribution('groclient').version
except DistributionNotFound:
    pass

class APIError(Exception):
    def __init__(self, response, retry_count, url, params):
        self.response = response
        self.retry_count = retry_count
        self.url = url
        self.params = params
        self.status_code = response.status_code if hasattr(response, 'status_code') else None
        try:
            json_content = self.response.json()
            # 'error' should be something like 'Not Found' or 'Bad Request'
            self.message = json_content.get('error', '')
            # Some error responses give additional info.
            # For example, a 400 Bad Request might say "metricId is required"
            if 'message' in json_content:
                self.message += ': {}'.format(json_content['message'])
        except Exception:
            # If the error message can't be parsed, fall back to a generic "giving up" message.
            self.message = 'Giving up on {} after {} {}: {}'.format(self.url, self.retry_count,
                                                                    'retry' if self.retry_count == 1
                                                                    else 'retries', response)


def get_default_logger():
    """Get a logging object using the default log level set in cfg.

    https://docs.python.org/3/library/logging.html

    Returns
    -------
    logger : logging.Logger

    """
    logger = logging.getLogger(__name__)
    if not logger.handlers:
        stderr_handler = logging.StreamHandler()
        logger.addHandler(stderr_handler)
    return logger


def get_access_token(api_host, user_email, user_password, logger=None):
    """Request an access token.

    Parameters
    ----------
    api_host : string
        The API host's url, excluding 'https://'
        ex. 'api.gro-intelligence.com'
    user_email : string
        Email address associated with user's Gro account
    user_password : string
        Password for user's Gro account
    logger : logging.Logger
        Alternative logger object if wanting to use a non-default one.
        Otherwise get_default_logger() will be used.

    Returns
    -------
    accessToken : string

    """
    retry_count = 0
    if not logger:
        logger = get_default_logger()
    while retry_count <= cfg.MAX_RETRIES:
        get_api_token = requests.post('https://' + api_host + '/api-token',
                                      data={'email': user_email,
                                            'password': user_password})
        if get_api_token.status_code == 200:
            logger.debug('Authentication succeeded in get_access_token')
            return get_api_token.json()['data']['accessToken']
        else:
            logger.warning('Error in get_access_token: {}'.format(get_api_token))
        retry_count += 1
    raise Exception('Giving up on get_access_token after {0} tries.'.format(retry_count))


def redirect(old_params, migration):
    """Update query parameters to follow a redirection response from the API.

    >>> redirect(
    ...     {'metricId': 14, 'sourceId': 2, 'itemId': 145},
    ...     {'old_metric_id': 14, 'new_metric_id': 15, 'source_id': 2}
    ... ) == {'sourceId': 2, 'itemId': 145, 'metricId': 15}
    True

    Parameters
    ----------
    old_params : dict
        The original parameters provided to the API request
    migration : dict
        The body of the 301 response indicating which of the inputs have been
        migrated and what values they have been migrated to

    Returns
    -------
    new_params : dict
        The mutated params object with values replaced according to the
        redirection instructions provided by the API

    """
    new_params = old_params.copy()
    for migration_key in migration:
        split_mig_key = migration_key.split('_')
        if split_mig_key[0] == 'new':
            param_key = groclient.utils.str_snake_to_camel('_'.join([split_mig_key[1], 'id']))
            new_params[param_key] = migration[migration_key]
    return new_params


def get_version_info():
    return _VERSIONS.copy()


def convert_value(value, from_convert_factor, to_convert_factor):
    value_in_base_unit = (
        value * from_convert_factor.get("factor")
    ) + from_convert_factor.get("offset", 0)

    return float(
        value_in_base_unit - to_convert_factor.get("offset", 0)
    ) / to_convert_factor.get("factor")


def get_data(url, headers, params=None, logger=None):
    """General 'make api request' function.

    Assigns headers and builds in retries and logging.

    Parameters
    ----------
    url : string
    headers : dict
    params : dict
    logger : logging.Logger

    Returns
    -------
    data : list or dict

    """
    base_log_record = dict(route=url, params=params)
    retry_count = 0

    # append version info
    headers.update(get_version_info())

    if not logger:
        logger = get_default_logger()
        logger.debug(url)
        logger.debug(params)
    while retry_count <= cfg.MAX_RETRIES:
        start_time = time.time()
        try:
            response = requests.get(url, params=params, headers=headers, timeout=None)
        except Exception as e:
            response = e
        elapsed_time = time.time() - start_time
        status_code = response.status_code if hasattr(response, 'status_code') else None
        log_record = dict(base_log_record)
        log_record['elapsed_time_in_ms'] = 1000 * elapsed_time
        log_record['retry_count'] = retry_count
        log_record['status_code'] = status_code
        if status_code == 200:  # Success
            logger.debug('OK', extra=log_record)
            return response
        if status_code in [204, 206]:  # Success with a caveat - warning
            log_msg = {204: 'No Content', 206: 'Partial Content'}[status_code]
            logger.warning(log_msg, extra=log_record)
            return response
        log_record['tag'] = 'failed_gro_api_request'
        if retry_count < cfg.MAX_RETRIES:
            logger.warning(response.text if hasattr(response, 'text') else response,
                           extra=log_record)
        if status_code in [400, 401, 402, 404]:
            break  # Do not retry
        if status_code == 301:
            new_params = redirect(params, response.json()['data'][0])
            logger.warning('Redirecting {} to {}'.format(params, new_params), extra=log_record)
            params = new_params
        else:
            logger.warning('{}'.format(response), extra=log_record)
            if retry_count > 0:
                # Retry immediately on first failure.
                # Exponential backoff before retrying repeatedly failing requests.
                time.sleep(2 ** retry_count)
        retry_count += 1
    raise APIError(response, retry_count, url, params)


@memoize(maxsize=None)
def get_allowed_units(access_token, api_host, metric_id, item_id):
    url = '/'.join(['https:', '', api_host, 'v2/units/allowed'])
    headers = {'authorization': 'Bearer ' + access_token}
    params = {'metricIds': metric_id}
    if item_id:
        params['itemIds'] = item_id
    resp = get_data(url, headers, params)
    return [unit['id'] for unit in resp.json()['data']]


@memoize(maxsize=None)
def get_available(access_token, api_host, entity_type):
    url = '/'.join(['https:', '', api_host, 'v2', entity_type])
    headers = {'authorization': 'Bearer ' + access_token}
    resp = get_data(url, headers)
    return resp.json()['data']


def list_available(access_token, api_host, selected_entities):
    url = '/'.join(['https:', '', api_host, 'v2/entities/list'])
    headers = {'authorization': 'Bearer ' + access_token}
    params = dict([(groclient.utils.str_snake_to_camel(key), value)
                   for (key, value) in list(selected_entities.items())])
    resp = get_data(url, headers, params)
    try:
        return resp.json()['data']
    except KeyError:
        raise Exception(resp.text)


@memoize(maxsize=None)
def lookup_single(access_token, api_host, entity_type, entity_id):
    url = '/'.join(['https:', '', api_host, 'v2', entity_type])
    headers = {'authorization': 'Bearer ' + access_token}
    params = {'ids': [entity_id]}
    resp = get_data(url, headers, params)
    try:
        return resp.json()['data'].get(str(entity_id))
    except KeyError:
        raise Exception(resp.text)


def lookup_batch(access_token, api_host, entity_type, entity_ids):
    url = '/'.join(['https:', '', api_host, 'v2', entity_type])
    headers = {'authorization': 'Bearer ' + access_token}
    all_results = {}
    for id_batch in groclient.utils.list_chunk(entity_ids):
        params = {'ids': id_batch}
        resp = get_data(url, headers, params)
        result = resp.json()['data']
        for id_str in result.keys():
            all_results[id_str] = result[id_str]
    return all_results


def lookup(access_token, api_host, entity_type, entity_ids):
    try:  # Convert iterable types like numpy arrays or tuples into plain lists
        entity_ids = list(entity_ids)
        return lookup_batch(access_token, api_host, entity_type, entity_ids)
    except TypeError:  # Convert anything else, like strings or numpy integers, into plain integers
        entity_id = int(entity_ids)
        # If an integer is given, return only the dict with that id
        return lookup_single(access_token, api_host, entity_type, entity_id)


def get_params_from_selection(**selection):
    """Construct http request params from dict of entity selections.

    For use with get_data_series() and rank_series_by_source().

    >>> get_params_from_selection(
    ...     metric_id=123, item_id=456, unit_id=14
    ... ) == { 'itemId': 456, 'metricId': 123 }
    True

    Parameters
    ----------
    metric_id : integer, optional
    item_id : integer, optional
    region_id : integer, optional
    partner_region_id : integer, optional
    source_id : integer, optional
    frequency_id : integer, optional
    start_date: string, optional
    end_date: string, optional

    Returns
    -------
    dict
        selections with valid keys converted to camelcase and invalid ones filtered out

    """
    params = {}
    for key, value in list(selection.items()):
        if key in ('region_id', 'partner_region_id', 'item_id', 'metric_id',
                   'source_id', 'frequency_id', 'start_date', 'end_date'):
            params[groclient.utils.str_snake_to_camel(key)] = value
    return params


def get_data_call_params(**selection):
    """Construct http request params from dict of entity selections.

    For use with get_data_points().

    >>> get_data_call_params(
    ...     metric_id=123, start_date='2012-01-01', unit_id=14
    ... ) == {'startDate': '2012-01-01', 'metricId': 123, 'responseType': 'list_of_series'}
    True

    Parameters
    ----------
    metric_id : integer
    item_id : integer
    region_id : integer
    partner_region_id : integer
    source_id : integer
    frequency_id : integer
    start_date : string, optional
    end_date : string, optional
    show_revisions : boolean, optional
    insert_null : boolean, optional
    show_metadata : boolean, optional
    at_time : string, optional

    Returns
    -------
    dict
        selections with valid keys converted to camelcase and invalid ones filtered out

    """
    params = get_params_from_selection(**selection)
    for key, value in list(selection.items()):
        if key == 'show_metadata':
            params[groclient.utils.str_snake_to_camel('show_meta_data')] = value
        if key in ('start_date', 'end_date', 'show_revisions', 'insert_null', 'at_time'):
            params[groclient.utils.str_snake_to_camel(key)] = value
    params['responseType'] = 'list_of_series'
    return params


def get_data_series(access_token, api_host, **selection):
    logger = get_default_logger()
    url = '/'.join(['https:', '', api_host, 'v2/data_series/list'])
    headers = {'authorization': 'Bearer ' + access_token}
    params = get_params_from_selection(**selection)
    resp = get_data(url, headers, params)
    try:
        response = resp.json()['data']
        if any((series.get('metadata', {}).get('includes_historical_region', False))
                for series in response):
            logger.warning('Data series have some historical regions, '
                           'see https://developers.gro-intelligence.com/faq.html')
        return response
    except KeyError:
        raise Exception(resp.text)


def get_top(access_token, api_host, entity_type, num_results=5, **selection):
    url = '/'.join(['https:', '', api_host, 'v2/top/{}'.format(entity_type)])
    headers = {'authorization': 'Bearer ' + access_token}
    params = get_params_from_selection(**selection)
    params['n'] = num_results
    resp = get_data(url, headers, params)
    try:
        return resp.json()
    except KeyError:
        raise Exception(resp.text)


def make_key(key):
    if key not in ('startDate', 'endDate'):
        return key + 's'
    return key


def get_source_ranking(access_token, api_host, series):
    """Given a series, return a list of ranked sources.

    :param access_token: API access token.
    :param api_host: API host.
    :param series: Series to calculate source raking for.
    :return: List of sources that match the series parameters, sorted by rank.
    """
    params = dict((make_key(k), v) for k, v in iter(list(
        get_params_from_selection(**series).items())))
    url = '/'.join(['https:', '', api_host, 'v2/available/sources'])
    headers = {'authorization': 'Bearer ' + access_token}
    return get_data(url, headers, params).json()


def rank_series_by_source(access_token, api_host, selections_list):
    series_map = OrderedDict()
    for selection in selections_list:
        series_key = '.'.join([json.dumps(selection.get(type_id))
                               for type_id in DATA_SERIES_UNIQUE_TYPES_ID
                               if type_id != 'source_id'])
        if series_key not in series_map:
            series_map[series_key] = {}
        elif None in series_map[series_key]:
            continue
        series_map[series_key][selection.get('source_id')] = selection

    for series_key, series_by_source_id in series_map.items():
        series_without_source = {
            type_id: json.loads(series_key.split('.')[idx])
            for idx, type_id in enumerate(DATA_SERIES_UNIQUE_TYPES_ID)
            if type_id != 'source_id' and series_key.split('.')[idx] != 'null'
        }
        try:
            source_ids = get_source_ranking(access_token,
                                            api_host,
                                            series_without_source)
        # Catch "no content" response from get_source_ranking()
        except ValueError:
            continue  # empty response

        for source_id in source_ids:
            if source_id in series_by_source_id:
                yield series_by_source_id[source_id]
            if None in series_by_source_id:
                yield groclient.utils.dict_assign(series_without_source, 'source_id', source_id)


def get_available_timefrequency(access_token, api_host, **series):
    params = dict((make_key(k), v) for k, v in iter(list(
        get_params_from_selection(**series).items())))
    url = '/'.join(['https:', '', api_host, 'v2/available/time-frequencies'])
    headers = {'authorization': 'Bearer ' + access_token}
    response = get_data(url, headers, params)
    if response.status_code == 204:
        return []
    return [groclient.utils.dict_reformat_keys(tf, groclient.utils.str_camel_to_snake)
            for tf in response.json()]


def list_of_series_to_single_series(series_list, add_belongs_to=False, include_historical=True):
    """Convert list_of_series format from API back into the familiar single_series output format."""
    if not isinstance(series_list, list):
        # If the output is an error or None or something else that's not a list, just propagate
        return series_list
    output = []
    for series in series_list:
        if not (isinstance(series, dict) and isinstance(series.get('data', []), list)):
            continue
        series_metadata = series.get('series', {}).get('metadata', {})
        has_historical_regions = (series_metadata.get('includesHistoricalRegion', False) or
                                  series_metadata.get('includesHistoricalPartnerRegion', False))
        if not include_historical and has_historical_regions:
            continue
        # All the belongsTo keys are in camelCase. Convert them to snake_case.
        # Only need to do this once per series, so do this outside of the list
        # comprehension and save to a variable to avoid duplicate work:
        belongs_to = groclient.utils.dict_reformat_keys(series.get('series', {}).get('belongsTo', {}),
                                                        groclient.utils.str_camel_to_snake)
        for point in series.get('data', []):
            formatted_point = {
                'start_date': point[0],
                'end_date': point[1],
                'value': point[2],
                'unit_id': point[4] if len(point) > 4 else series['series'].get('unitId', None),
                'metadata': point[5] if len(point) > 5 else {},
                # input_unit_id and input_unit_scale are deprecated but provided for backwards
                # compatibility. unit_id should be used instead.
                'input_unit_id': point[4] if len(point) > 4 else series['series'].get('unitId', None),
                'input_unit_scale': 1,
                # If a point does not have reporting_date, use None
                'reporting_date': point[3] if len(point) > 3 else None,
                # Series attributes:
                'metric_id': series['series'].get('metricId', None),
                'item_id': series['series'].get('itemId', None),
                'region_id': series['series'].get('regionId', None),
                'partner_region_id': series['series'].get('partnerRegionId', 0),
                'frequency_id': series['series'].get('frequencyId', None)
                # 'source_id': series['series'].get('sourceId', None), TODO: add source to output
            }
            if formatted_point['metadata'].get('confInterval') is not None:
                formatted_point['metadata']['conf_interval'] = formatted_point['metadata'].pop('confInterval')

            if add_belongs_to:
                # belongs_to is consistent with the series the user requested. So if an
                # expansion happened on the server side, the user can reconstruct what
                # results came from which request.
                formatted_point['belongs_to'] = belongs_to
            output.append(formatted_point)
    return output


def get_data_points(access_token, api_host, **selection):
    headers = {'authorization': 'Bearer ' + access_token}
    url = '/'.join(['https:', '', api_host, 'v2/data'])
    params = get_data_call_params(**selection)
    resp = get_data(url, headers, params)
    include_historical = selection.get('include_historical', True)
    return list_of_series_to_single_series(resp.json(), False, include_historical)


@memoize(maxsize=None)
def universal_search(access_token, api_host, search_terms):
    """Search across all entity types for the given terms.

    Parameters
    ----------
    access_token : string
    api_host : string
    search_terms : string

    Returns
    -------
    list of [id, entity_type] pairs

        Example::

            [[5604, 'item'], [10204, 'item'], [410032, 'metric'], ....]

    """
    url_pieces = ['https:', '', api_host, 'v2/search']
    url = '/'.join(url_pieces)
    headers = {'authorization': 'Bearer ' + access_token}
    resp = get_data(url, headers, {'q': search_terms})
    return resp.json()


@memoize(maxsize=None)
def search(access_token, api_host, entity_type, search_terms):
    url = '/'.join(['https:', '', api_host, 'v2/search', entity_type])
    headers = {'authorization': 'Bearer ' + access_token}
    resp = get_data(url, headers, {'q': search_terms})
    return resp.json()


def search_and_lookup(access_token, api_host, entity_type, search_terms, num_results=10):
    search_results = search(access_token, api_host, entity_type, search_terms)[:num_results]
    search_result_ids = [result['id'] for result in search_results]
    search_result_details = lookup(access_token, api_host, entity_type, search_result_ids)
    for search_result_id in search_result_ids:
        yield search_result_details[str(search_result_id)]


def lookup_belongs(access_token, api_host, entity_type, entity_id):
    parent_ids = lookup(access_token, api_host, entity_type, entity_id)['belongsTo']
    parent_details = lookup(access_token, api_host, entity_type, parent_ids)
    for parent_id in parent_ids:
        yield parent_details[str(parent_id)]


def get_geo_centre(access_token, api_host, region_id):
    url = '/'.join(['https:', '', api_host, 'v2/geocentres'])
    headers = {'authorization': 'Bearer ' + access_token}
    resp = get_data(url, headers, {'regionIds': region_id})
    return resp.json()['data']


@memoize(maxsize=None)
def get_geojsons(access_token, api_host, region_id, descendant_level, zoom_level):
    url = '/'.join(['https:', '', api_host, 'v2/geocentres'])
    params = {'includeGeojson': True, 'regionIds': region_id, 'zoom': zoom_level}
    if descendant_level:
        params['reqRegionLevelId']= descendant_level
        params['stringify'] = 'false'
    headers = {'authorization': 'Bearer ' + access_token}
    resp = get_data(url, headers, params)
    return [groclient.utils.dict_reformat_keys(r, groclient.utils.str_camel_to_snake)
            for r in resp.json()['data']]


def get_geojson(access_token, api_host, region_id, zoom_level):
    for region in get_geojsons(access_token, api_host, region_id, None, zoom_level):
        return json.loads(region['geojson'])


def get_descendant(access_token, api_host, entity_type, entity_id, distance=None,
                   include_details=True):
    url = '/'.join(['https:', '', api_host, 'v2/{}/contains'.format(entity_type)])
    headers = {'authorization': 'Bearer ' + access_token}
    params = {'ids': [entity_id]}
    if distance:
        params['distance'] = distance
    else:
        params['distance'] = -1

    resp = get_data(url, headers, params)
    descendant_entity_ids = resp.json()['data'][str(entity_id)]

    if include_details:
        entity_details = lookup(access_token, api_host, entity_type, descendant_entity_ids)
        return [entity_details[str(child_entity_id)] for child_entity_id in descendant_entity_ids]

    return [{'id': descendant_entity_id} for descendant_entity_id in descendant_entity_ids]


def get_descendant_regions(access_token, api_host, region_id,
                           descendant_level=None, include_historical=True, include_details=True):
    url = '/'.join(['https:', '', api_host, 'v2/regions/contains'])
    headers = {'authorization': 'Bearer ' + access_token}
    params = {'ids': [region_id]}
    if descendant_level:
        params['level'] = descendant_level
    else:
        params['distance'] = -1

    resp = get_data(url, headers, params)
    descendant_region_ids = resp.json()['data'][str(region_id)]

    # Filter out regions with the 'historical' flag set to true
    if not include_historical or include_details:
        region_details = lookup(access_token, api_host, 'regions', descendant_region_ids)

        if not include_historical:
            descendant_region_ids = [region['id'] for region in region_details.values()
                                     if not region['historical']]

        if include_details:
            return [region_details[str(region_id)] for region_id in descendant_region_ids]

    return [{'id': descendant_region_id} for descendant_region_id in descendant_region_ids]


if __name__ == '__main__':
    # To run doctests:
    # $ python lib.py -v
    import doctest
    doctest.testmod(raise_on_error=True,  # Set to False for prettier error message
                    optionflags=doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS)



try:
    # Python 3.3+
    from unittest.mock import patch, MagicMock
    from io import StringIO
except ImportError:
    # Python 2.7
    from mock import patch, MagicMock
    from StringIO import StringIO
from unittest import TestCase
import json
from datetime import date

from tornado.httpclient import HTTPResponse, HTTPError
from tornado.concurrent import Future
from tornado.ioloop import IOLoop

from groclient import GroClient
from groclient.client import BatchError
from groclient.utils import str_camel_to_snake
from groclient.mock_data import (
    mock_list_of_series_points,
    mock_data_series,
    mock_error_selection,
)

MOCK_HOST = "pytest.groclient.url"
MOCK_TOKEN = "pytest.groclient.token"


def mock_rank_series_by_source(access_token, api_host, selections_list):
    for data_series in mock_data_series:
        yield data_series


def mock_tornado_fetch(request):
    future = Future()
    query_params = {
        str_camel_to_snake(key): value
        for key, value in [
            query_param.split("=")
            for query_param in request.url.split("?")[1].split("&")
        ]
    }
    if int(query_params.get("item_id", 0)) < 0:
        raise HTTPError(400, "Negative item ids are not allowed", request)
    else:
        response = HTTPResponse(
            request, 200, buffer=StringIO(json.dumps(mock_list_of_series_points))
        )
    future.set_result(response)
    return future


@patch(
    "groclient.lib.rank_series_by_source",
    MagicMock(side_effect=mock_rank_series_by_source),
)
@patch(
    "tornado.httpclient.AsyncHTTPClient.fetch",
    MagicMock(side_effect=mock_tornado_fetch),
)
class BatchTests(TestCase):
    def setUp(self):
        self.client = GroClient(MOCK_HOST, MOCK_TOKEN)
        self.assertTrue(isinstance(self.client, GroClient))

    def tearDown(self):
        IOLoop.clear_current()

    def test_batch_async_get_data_points(self):
        data_points = self.client.batch_async_get_data_points(
            [
                {
                    "metric_id": 1,
                    "item_id": 2,
                    "region_id": 3,
                    "frequency_id": 4,
                    "source_id": 5,
                },
                {
                    "metric_id": 6,
                    "item_id": 7,
                    "region_id": 8,
                    "frequency_id": 9,
                    "source_id": 10,
                    "insert_nulls": True,
                },
            ]
        )
        self.assertEqual(data_points[0][0]["start_date"], "2017-01-01T00:00:00.000Z")
        self.assertEqual(data_points[0][0]["end_date"], "2017-12-31T00:00:00.000Z")
        self.assertEqual(data_points[0][0]["value"], 40891)
        self.assertEqual(data_points[0][0]["unit_id"], 14)
        self.assertEqual(data_points[0][0]["reporting_date"], None)
        self.assertEqual(data_points[0][1]["start_date"], "2018-01-01T00:00:00.000Z")
        self.assertEqual(data_points[0][1]["end_date"], "2018-12-31T00:00:00.000Z")
        self.assertEqual(data_points[0][1]["value"], 56789)
        self.assertEqual(data_points[0][1]["unit_id"], 10)
        self.assertEqual(
            data_points[0][1]["reporting_date"], "2019-03-14T00:00:00.000Z"
        )

    def test_batch_async_get_data_points_map_function(self):
        def sum_results(inputIndex, inputObject, response, summation):
            for point in response:
                summation += point["value"]
            return summation

        summation = self.client.batch_async_get_data_points(
            [
                {
                    "metric_id": 1,
                    "item_id": 2,
                    "region_id": 3,
                    "frequency_id": 4,
                    "source_id": 5,
                }
            ],
            output_list=0,
            map_result=sum_results,
        )

        self.assertEqual(summation, 97680)

    def test_batch_async_get_data_points_bad_request_error(self):
        responses = self.client.batch_async_get_data_points([mock_error_selection])
        self.assertTrue(isinstance(responses[0], BatchError))

    def test_batch_async_get_data_points_map_errors(self):
        def raise_exception(idx, query, response, accumulator):
            if isinstance(response, Exception):
                raise response
            accumulator[idx] = response
            return accumulator

        with self.assertRaises(Exception):
            self.client.batch_async_get_data_points(
                [mock_error_selection], map_result=raise_exception
            )

    def test_async_get_df(self):
        self.client.add_single_data_series(
            {
                "metric_id": 1,
                "item_id": 2,
                "region_id": 3,
                "frequency_id": 4,
                "source_id": 5,
            }
        )
        df = self.client.async_get_df()
        self.assertEqual(df.iloc[0]["start_date"].date(), date(2017, 1, 1))
        self.assertEqual(df.iloc[0]["end_date"].date(), date(2017, 12, 31))
        self.assertEqual(df.iloc[0]["value"], 40891)

    def test_batch_async_rank_series_by_source(self):
        list_of_ranked_series_lists = self.client.batch_async_rank_series_by_source(
            [mock_data_series, mock_data_series]
        )
        # There were two inputs, so there should be two outputs:
        self.assertEqual(len(list_of_ranked_series_lists), 2)
        for series_list in list_of_ranked_series_lists:
            # Not necessarily true, but true given the mock_rank_series_by_source() function:
            self.assertEqual(len(series_list), len(mock_data_series))
            for series in series_list:
                self.assertTrue("metric_id" in series)
                self.assertTrue("source_id" in series)



from unittest import TestCase

from groclient.utils import (
    str_camel_to_snake,
    str_snake_to_camel,
    dict_reformat_keys,
    dict_unnest,
    list_chunk,
    intersect,
    zip_selections,
)


class UtilsTests(TestCase):
    def test_str_camel_to_snake(self):
        self.assertEqual(str_camel_to_snake("partnerRegionId"), "partner_region_id")

        self.assertEqual(str_camel_to_snake("partner_region_id"), "partner_region_id")

    def test_str_snake_to_camel(self):
        self.assertEqual(str_snake_to_camel("hello_world"), "helloWorld")

    def test_dict_reformat_keys(self):
        self.assertEqual(
            dict_reformat_keys({"belongsTo": {"metricId": 4}}, str_camel_to_snake),
            {"belongs_to": {"metricId": 4}},
        )

        self.assertEqual(
            dict_reformat_keys({"belongs_to": {"metric_id": 4}}, str_snake_to_camel),
            {"belongsTo": {"metric_id": 4}},
        )

    def test_dict_unnest(self):
        self.assertEqual(
            dict_unnest({"metric_id": 14, "belongs_to": {"metric_id": 14}}),
            {"metric_id": 14, "belongs_to_metric_id": 14},
        )

        self.assertEqual(
            dict_unnest(
                {
                    "metric_id": 14,
                    "belongs_to": {
                        "metric_id": 14,
                        "metadata": {"includes_historical": True},
                    },
                }
            ),
            {
                "metric_id": 14,
                "belongs_to_metric_id": 14,
                "belongs_to_metadata_includes_historical": True,
            },
        )

    def test_list_chunk(self):
        self.assertEqual(
            list_chunk([1, 2, 3, 4, 5, 6, 7, 8], 3), [[1, 2, 3], [4, 5, 6], [7, 8]]
        )
        self.assertEqual(
            list_chunk([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11], 5),
            [[1, 2, 3, 4, 5], [6, 7, 8, 9, 10], [11]],
        )

    def test_intersect(self):
        self.assertEqual(intersect([1, 2, 3], [4, 5, 6]), [])
        self.assertEqual(intersect([1, 2, 3], [4, 5, 6, 2]), [2])
        self.assertEqual(intersect([1, 2, 3], [1, 2, 3]), [1, 2, 3])

    def test_zip_selections(self):
        self.assertEqual(
            zip_selections([860032, 274, 1215, 0, 2, 9]),
            {
                "metric_id": 860032,
                "item_id": 274,
                "region_id": 1215,
                "partner_region_id": 0,
                "frequency_id": 2,
                "source_id": 9,
            },
        )
        self.assertEqual(
            zip_selections(
                [860032, 274, 1215, 0, 2, 9], insert_nulls=True, metric_id=1
            ),
            {
                "metric_id": 860032,
                "item_id": 274,
                "region_id": 1215,
                "partner_region_id": 0,
                "frequency_id": 2,
                "source_id": 9,
                "insert_nulls": True,
            },
        )



"""
Numpy data transforms for API data.

This file contains a library of Transformer classes that expand sklearn transformation functionality
http://scikit-learn.org/stable/data_transforms.html, and necessary utilities.

Each transformer class has a fit method (generic in most cases) and a transform method.
The transform methods return a transformed numpy array, sometimes with a list of features,
if the transformed dataframe has different columns as the original one.

Eventually these could be incorporated into the API to reduce client side complexity for customers.
"""

import dateparser
import numpy as np
from sklearn.base import TransformerMixin
from scipy.fftpack import fft
from datetime import datetime, timedelta

class Transformer(TransformerMixin):
    """
    Base class of tranformers
    """
    def __init__(self):
        self.features = 0 #number of output features or desired number of features

    def transform(self, input):
        """
        Return a transformed version of the input. No guarantees are made about the format or dimensions of the output.
        :param input: a numpy array
        :return: a transformed version of the numpy array
        """
        pass

    def get_num_features(self):
        """
        Return the number of features this transformer will give for a given input size.
        :return: number of features in the returned data point.
        """
        pass


class FourierCoef(Transformer):

    def __init__(self, start_idx, num_features):

        """
        Initialize the fourier transform.
        :param features: Number of coefficients to output. Gives the n lowest frequency coefficients.
        pass -1 for all coefficients.
        """

        self.num_features = num_features
        self.start_idx = start_idx

    def transform(self, time_series):
        """
        Return the coefficients of the data.
        :param time_series: the timeseries values sampled at a regular interval with sampling frequency 1 we assume.
        :return: The coefficients after the linear shift.
        """

        # FFT by default computes over the last axis.
        ranges = np.r_[0,self.start_idx:self.start_idx + self.num_features - 1].astype(int)
        assert(len(ranges) == self.num_features)
        fft_time_series = fft(time_series)[ranges]

        #Only interested in magnitude! not in phase in this case
        return np.abs(fft_time_series)

# Other transformers
def post_process_timeseries(num_of_points, start_datetime, time_series, start_idx, num_features, period_length_days=1):
    dataset_imputed, coverage = _impute(num_of_points, period_length_days, start_datetime, time_series)

    # if imputing fails, we can't do fft either
    if dataset_imputed is None:
        return None, 0.0

    fourier = FourierCoef(start_idx, num_features)
    coefs = fourier.transform(dataset_imputed)

    return coefs, coverage

# temporary until filling nulls is enabled in prod api
def _fill_in_blank_days(num_of_points, start_datetime, pulled_dataset):
    dataset = np.zeros(num_of_points)

    ptr_idx_incomplete_list = 0

    their_start_date = dateparser.parse(pulled_dataset[0]["start_date"])

    date_delta = 0

    if their_start_date.date() < start_datetime.date():
        date_delta = (start_datetime.date() - their_start_date.date()).days

    ptr_idx_incomplete_list += date_delta

    for day_idx in range(0, num_of_points):
        curr_idx_date = (start_datetime + timedelta(days=day_idx)).isoformat() + ".000Z"

        if ptr_idx_incomplete_list >= len(pulled_dataset) or pulled_dataset[ptr_idx_incomplete_list][
            "start_date"] != curr_idx_date:
            # add null entry at this point for later imputation.
            dataset[day_idx] = float('NaN')
        else:
            dataset[day_idx] = pulled_dataset[ptr_idx_incomplete_list]["value"]
            ptr_idx_incomplete_list += 1

    return dataset

def _impute(num_of_points, period_length_days, start_datetime, pulled_dataset):
    x_output = np.arange(0, period_length_days * num_of_points, period_length_days)

    x_input = []
    y_input = []

    num_of_valid_points_pulled_dataset = 0

    for datapoint in pulled_dataset:
        if datapoint["value"] is not None:
            delta = (datetime.strptime(datapoint["start_date"][0:10], "%Y-%m-%d").date() - start_datetime.date()).days
            x_input.append(delta)
            y_input.append(datapoint["value"])
            num_of_valid_points_pulled_dataset += 1

    # num data points to interpolate
    if len(x_input) == 0:
        return None, 0.0

    y_output = np.interp(x_output, x_input, y_input)
    coverage = float(num_of_valid_points_pulled_dataset) / num_of_points

    return y_output, coverage




from groclient import GroClient
import mock
import numpy as np
import pandas as pd
from pandas.util.testing import assert_frame_equal
import pytest

from api.client.samples.analogous_years.lib import get_transform_data


def create_test_data():
    return {'end_date': ['2019-07-31T00:00:00.000Z',
                         '2005-08-28T00:00:00.000Z',
                         '2000-03-01T00:00:00.000Z',
                         '2005-08-28T00:00:00.000Z'],
            'frequency_id': [1, 1, 1, 1],
            'input_unit_id': [2, 2, 2, 2],
            'input_unit_scale': [1, 1, 1, 1],
            'item_id': [2039, 2039, 2039, 2039],
            'metric_id': [2100031, 2100031, 2100031, 2100031],
            'region_id': [13100, 13065, 1107, 13064],
            'start_date': ['2019-07-31T00:00:00.000Z',
                           '2005-08-28T00:00:00.000Z',
                           '2000-03-01T00:00:00.000Z',
                           '2005-08-28T00:00:00.000Z'],
            'unit_id': [2, 2, 2, 2],
            'value': [0.13002748115958, 1.17640700229636,
                      2.39664378851418, 1.10551943121531]}


def create_test_data_for_get_data():
    return pd.DataFrame({'end_date': pd.to_datetime(['2019-07-31T00:00:00.000Z',
                                                     '2019-08-03T00:00:00.000Z',
                                                     '2019-08-05T00:00:00.000Z',
                                                     '2019-08-10T00:00:00.000Z']),
                         'frequency_id': [1, 1, 1, 1],
                         'input_unit_id': [2, 2, 2, 2],
                         'input_unit_scale': [1, 1, 1, 1],
                         'item_id': [2039, 2039, 2039, 2039],
                         'metric_id': [2100031, 2100031, 2100031, 2100031],
                         'region_id': [1215, 1215, 1215, 1215],
                         'start_date': pd.to_datetime(['2019-07-31T00:00:00.000Z',
                                                       '2019-08-03T00:00:00.000Z',
                                                       '2019-08-05T00:00:00.000Z',
                                                       '2019-08-10T00:00:00.000Z']),
                         'unit_id': [2, 2, 2, 2],
                         'value': [0.13002748115958, 1.17640700229636,
                                   2.39664378851418, 1.10551943121531]})


@mock.patch('api.client.gro_client.GroClient.get_df', return_value=create_test_data_for_get_data())
def test_get_data(test_data_1):
    client = GroClient('mock_website', 'mock_access_token')
    start_date_bound = '2019-08-01T00:00:00.000Z'
    expected = pd.DataFrame(pd.DataFrame({'end_date': pd.to_datetime(['2019-08-01T00:00:00.000Z',
                                                                      '2019-08-02T00:00:00.000Z',
                                                                      '2019-08-03T00:00:00.000Z',
                                                                      '2019-08-04T00:00:00.000Z',
                                                                      '2019-08-05T00:00:00.000Z',
                                                                      '2019-08-06T00:00:00.000Z',
                                                                      '2019-08-07T00:00:00.000Z',
                                                                      '2019-08-08T00:00:00.000Z',
                                                                      '2019-08-09T00:00:00.000Z',
                                                                      '2019-08-10T00:00:00.000Z']),
                                          'value': [0.13002748115958, 1.17640700229636,
                                                    1.17640700229636, 2.39664378851418,
                                                    2.39664378851418, 2.39664378851418,
                                                    2.39664378851418, 1.10551943121531,
                                                    1.10551943121531, 1.10551943121531]}))
    expected.index = expected['end_date']
    expected = expected.asfreq('D')

    test_data = get_transform_data.get_data(client, 'metric_id', 'item_id', 'region_id',
                                            'source_id', 'frequency_id', start_date_bound)
    assert_frame_equal(test_data, expected)


def test_combine_subregions_with_subregion():
    test_data = pd.DataFrame(create_test_data())
    test_data.loc[:, 'end_date'] = pd.to_datetime(test_data['end_date'])
    test_data_subregion = test_data[['end_date', 'value']]
    expected_subregion = pd.DataFrame({'end_date': ['2000-03-01T00:00:00.000Z',
                                                    '2005-08-28T00:00:00.000Z',
                                                    '2019-07-31T00:00:00.000Z'],
                                       'value': [2.39664378851418,
                                                 2.28192643351167,
                                                 0.13002748115958]})
    utc_tz = False
    if pd.to_datetime(expected_subregion['end_date'][0]).tzinfo:
        utc_tz = True
    # Depending on the version of pandas, the data type of the datetime object
    # can vary between 'datetime64[ns, UTC]' and 'datetime64[ns]'
    expected_subregion.loc[:, 'end_date'] = pd.to_datetime(expected_subregion['end_date'],
                                                           utc=utc_tz)
    # The order of the columns after applying consolidation function is the following
    expected_subregion = expected_subregion[['value', 'end_date']]
    # The dataframes index columns is same as 'end_date' column
    expected_subregion.index = expected_subregion['end_date']
    # Test the equality of frames
    # expected_subregion.index = pd.to_datetime(expected_subregion.index)
    expected_subregion.loc[:, 'end_date'] = expected_subregion.index
    assert_frame_equal(get_transform_data.combine_subregions(test_data_subregion),
                       expected_subregion)


def test_combine_subregions_with_nosubregion():
    test_data = pd.DataFrame(create_test_data())
    test_data.loc[:, 'end_date'] = pd.to_datetime(test_data['end_date'])
    test_data_subregion = test_data[['end_date', 'value']]
    test_data_nosubregion = test_data_subregion.drop(test_data_subregion.index[-1])
    expected_nosubregion = pd.DataFrame({'end_date': ['2019-07-31T00:00:00.000Z',
                                                      '2005-08-28T00:00:00.000Z',
                                                      '2000-03-01T00:00:00.000Z'],
                                         'value': [0.13002748115958,
                                                   1.17640700229636,
                                                   2.39664378851418]})
    utc_tz = False
    if pd.to_datetime(expected_nosubregion['end_date'][0]).tzinfo:
        utc_tz = True
    # Depending on the version of pandas, the data type of the datetime object
    # can vary between 'datetime64[ns, UTC]' and 'datetime64[ns]'
    expected_nosubregion.index = expected_nosubregion['end_date']
    expected_nosubregion.loc[:, 'end_date'] = pd.to_datetime(expected_nosubregion['end_date'],
                                                             utc=utc_tz)
    expected_nosubregion.index = pd.to_datetime(expected_nosubregion.index, utc=utc_tz)
    expected_nosubregion.loc[:, 'end_date'] = pd.to_datetime(expected_nosubregion.index, utc=utc_tz)
    assert_frame_equal(get_transform_data.combine_subregions(test_data_nosubregion),
                       expected_nosubregion)


def test_combine_subregions_nodate():
    test_data = pd.DataFrame(create_test_data())
    test_data.loc[:, 'end_date'] = pd.to_datetime(test_data['end_date'])
    test_data_subregion = test_data[['end_date', 'value']]
    test_data_no_date = test_data_subregion.drop('end_date', axis=1)
    with pytest.raises(KeyError):
        get_transform_data.combine_subregions(test_data_no_date)


def test_extract_time_periods_by_dates_invalid_test_dates():
    test_data = pd.DataFrame(create_test_data())
    test_data.loc[:, 'end_date'] = pd.to_datetime(test_data['end_date'])
    # this is an invalid date because the initial and final dates are not within 1 year
    invalid_initial_date = '2016-02-02'
    invalid_final_date = '2017-02-02'
    with pytest.raises(ValueError):
        get_transform_data.extract_time_periods_by_dates(test_data,
                                                         invalid_initial_date,
                                                         invalid_final_date)


def test_extract_time_periods_by_dates_invalid_test_data():
    test_data = pd.DataFrame(create_test_data())
    # this is invalid test data as there is no 'end_date' column
    invalid_test_data = test_data.drop('end_date', axis=1)
    initial_date = '2016-02-02'
    final_date = '2017-02-01'
    with pytest.raises(KeyError):
        get_transform_data.extract_time_periods_by_dates(invalid_test_data,
                                                         initial_date,
                                                         final_date)


def test_extract_time_periods_by_dates_with_non_unique_dates():
    test_data = pd.DataFrame(create_test_data())
    initial_date = '2019-03-01'
    final_date = '2019-07-31'
    expected_data = test_data
    expected_data.loc[:, 'date'] = pd.to_datetime(expected_data['end_date'])
    expected_data.loc[:, 'year'] = expected_data['date'].dt.year.astype('str')
    expected_data.loc[:, 'period'] = ['2019-03-01 to 2019-07-31',
                                      '2005-03-01 to 2005-07-31',
                                      '2000-03-01 to 2000-07-31',
                                      '2005-03-01 to 2005-07-31']
    expected_data = expected_data[expected_data.period != '2005-03-01 to 2005-07-31']
    expected_data.loc[:, 'mm-dd'] = expected_data['date'].dt.strftime("%m-%d")
    expected_data = expected_data.sort_values(by=list(expected_data.columns), axis=0,
                                              ascending=False)
    assert_frame_equal(
        get_transform_data.extract_time_periods_by_dates(test_data,
                                                         initial_date,
                                                         final_date),
        expected_data)


def test_stack_time_periods_by_ddmm_nonunique_dates():
    test_data_invalid = pd.DataFrame(create_test_data())
    test_data_invalid.loc[:, 'end_date'] = pd.to_datetime(test_data_invalid['end_date'])
    with pytest.raises(KeyError):
        get_transform_data.stack_time_periods_by_ddmm(test_data_invalid)


def test_stack_time_periods_by_ddmm_unique_dates():
    test_data_invalid = pd.DataFrame(create_test_data())
    test_data_invalid.loc[:, 'end_date'] = pd.to_datetime(test_data_invalid['end_date'])
    test_data_invalid.loc[:, 'date'] = pd.to_datetime(test_data_invalid['end_date'])
    test_data_invalid.loc[:, 'period'] = [
        '2019-01-01 to 2019-12-31',
        '2005-01-01 to 2005-12-31',
        '2000-01-01 to 2000-12-31',
        '2005-01-01 to 2005-12-31']
    test_data_invalid.loc[:, 'mm-dd'] = test_data_invalid['date'].dt.strftime("%m-%d")
    test_data = test_data_invalid.drop(test_data_invalid.index[-1])
    expected = pd.DataFrame({'mm-dd': ['03-01', '07-31', '08-28'],
                             '2000-01-01 to 2000-12-31': [2.39664378851418,
                                                          np.nan,
                                                          np.nan],
                             '2005-01-01 to 2005-12-31': [np.nan,
                                                          np.nan,
                                                          1.17640700229636],
                             '2019-01-01 to 2019-12-31': [np.nan,
                                                          0.13002748115958,
                                                          np.nan]}).set_index('mm-dd')
    expected.columns.name = 'period'
    assert_frame_equal(get_transform_data.stack_time_periods_by_ddmm(test_data), expected)


def test_dates_to_period_string():
    test_dates = ['2019-07-31T00:00:00.000Z',
                  '2019-08-28T00:00:00.000Z']
    test_dates = pd.to_datetime(test_dates)
    expected = '2019-07-31 to 2019-08-28'
    assert get_transform_data.dates_to_period_string(test_dates[0], test_dates[1]) == expected


def test_loop_initiation_dates_invalid_initial_greater_than_final():
    test_max_date = pd.to_datetime('2019-08-31T00:00:00.000Z')
    invalid_initial_date = '2019-07-31'
    invalid_final_date = '2019-06-30'
    with pytest.raises(ValueError):
        get_transform_data.loop_initiation_dates(
            test_max_date, invalid_initial_date, invalid_final_date)


def test_loop_initiation_dates_invalid_final_greater_than_year_from_initial():
    test_max_date = pd.to_datetime('2019-08-31T00:00:00.000Z')
    invalid_initial_date = '2018-06-30'
    invalid_final_date = '2019-07-31'
    with pytest.raises(ValueError):
        get_transform_data.loop_initiation_dates(
            test_max_date, invalid_initial_date, invalid_final_date)


def test_loop_initiation_dates_invalid_final_greater_than_max():
    test_max_date = pd.to_datetime('2019-08-31T00:00:00.000Z')
    invalid_initial_date = '2019-07-31'
    invalid_final_date = '2019-09-30'
    with pytest.raises(ValueError):
        get_transform_data.loop_initiation_dates(
            test_max_date, invalid_initial_date, invalid_final_date)


def test_loop_initiation_dates():
    test_max_date = pd.to_datetime('2019-08-31T00:00:00.000Z')
    invalid_initial_date = '2016-12-31'
    invalid_final_date = '2017-12-01'
    # Depending on the version of pandas, the data type of the datetime object
    # can vary between 'datetime64[ns, UTC]' and 'datetime64[ns]'
    utc_tz = False
    if test_max_date.tzinfo:
        utc_tz = True
    expected = {'initial_date': pd.to_datetime('2017-12-31', utc=utc_tz),
                'final_date': pd.to_datetime('2018-12-01', utc=utc_tz)}
    assert get_transform_data.loop_initiation_dates(
        test_max_date, invalid_initial_date, invalid_final_date) == expected



import numpy as np
import unittest
from datetime import datetime
from api.client.samples.similar_regions.transform import _fill_in_blank_days, FourierCoef

class TransformTests(unittest.TestCase):

    def setUp(self):
        self.fourier = FourierCoef(0, 10)
        pass

    def test_fill_in_days(self):

        fake_pulled_dataset = [{u'value': 1, u'start_date': u'2018-09-23T00:00:00.000Z'},
                               {u'value': 2, u'start_date': u'2018-09-24T00:00:00.000Z'},
                               {u'value': 3, u'start_date': u'2018-09-25T00:00:00.000Z'},
                               {u'value': 4, u'start_date': u'2018-09-26T00:00:00.000Z'},
                               {u'value': 5, u'start_date': u'2018-09-28T00:00:00.000Z'},
                               {u'value': 6, u'start_date': u'2018-09-29T00:00:00.000Z'}]

        start_datetime = datetime(2018, 9, 25)

        np.testing.assert_equal(_fill_in_blank_days(10, start_datetime, fake_pulled_dataset),
                         np.array([3.0, 4.0, float('NaN'), 5.0, 6.0, float('NaN'),
                          float('NaN'), float('NaN'), float('NaN'), float('NaN')]))

        return

    def test_fourier(self):

        # we want to make sure we test phase independance here as well

        x = np.linspace(-np.pi, np.pi, 201)
        y = np.sin(5*x)
        coef = self.fourier.transform(y)

        y_with_phase = np.sin(5*x+0.2345)
        coef_with_phase = self.fourier.transform(y_with_phase)

        # the biggest coef should be the same...
        self.assertEqual(np.argmax(coef), np.argmax(coef_with_phase))

        # We shouldn't have big discrepancies
        np.testing.assert_allclose(coef, coef_with_phase, rtol=1e-0, atol=0)


if __name__ == '__main__':
    unittest.main()



"""
This library contains utilities for
1. computing pairwise distances between rows of a dataframe
2. selecting and sorting a specified column of the dataframe
"""

from dtw import dtw
import numpy as np
import pandas as pd
from scipy.spatial.distance import euclidean, pdist, squareform
from sklearn.preprocessing import MaxAbsScaler


def euclidean_dist_matrix(dataframe):
    """
    Calculates a square matrix of euclidean distances between the
    rows of the input dataframe
    :param dataframe: A pandas dataframe
    :return: A pandas dataframe
    """
    df = pd.DataFrame(squareform(pdist(dataframe)))
    df.index = dataframe.index
    df.columns = dataframe.index
    return df


def dtw_dist_matrix(dataframe):
    """
    Calculates a square matrix of dynamic time warping distances between the
    rows of the input dataframe
    :param dataframe: A pandas dataframe
    :return: A pandas dataframe
    """
    distance_matrix = np.zeros((len(dataframe.columns), len(dataframe.columns)))
    for i in range(len(dataframe.columns)):
        for j in range(i, len(dataframe.columns)):
            distance_matrix[i, j] = dtw(dataframe.iloc[:, i],
                                        dataframe.iloc[:, j],
                                        dist=euclidean)[0]
            distance_matrix[j, i] = dtw(dataframe.iloc[:, i],
                                        dataframe.iloc[:, j],
                                        dist=euclidean)[0]
    distance_matrix = pd.DataFrame(distance_matrix)
    distance_matrix.columns = dataframe.columns
    distance_matrix.index = dataframe.columns
    return distance_matrix


def scaled_labeled_method_distances(distance_matrix_df, initial_date, final_date, method):
    """
    Retrieves and scales a column of the input dataframe
    :param distance_matrix_df: A pandas dataframe
    :param final_date: 'YYYY-MM-DD'
    :param initial_date: 'YYYY-MM-DD'
    :param method: method used to compute the distance from the list
    ['cumulative' + GroItem, 'euclidean' + GroItem, 'dtw' + GroItem, 'tsfresh' + GroItem]
    :return: A pandas dataframe
    """
    list_of_methods = ['euclidean', 'cumulative', 'dtw', 'ts-features']
    if method.split('_')[0] not in list_of_methods:
        raise ValueError('Method of calculation unavailable')
    column_name = initial_date + ' to ' + final_date
    ranked_periods_df = pd.DataFrame(distance_matrix_df[column_name])
    scaler = MaxAbsScaler()
    ranked_periods_df.loc[:, column_name] = scaler.fit_transform(ranked_periods_df[[column_name]])
    ranked_periods_df.rename(columns={column_name: method}, inplace=True)
    return ranked_periods_df



from random import random
import argparse
import getpass
import os
import sys
import unicodecsv

import groclient.lib
import groclient.cfg
from groclient import GroClient


def pick_random_data_series(client):  # pragma: no cover
    """Pick a random available data series."""
    data_series_list = client.get_data_series()
    random_data_series = data_series_list[int(len(data_series_list) * random())]
    return random_data_series


def print_one_data_series(client, data_series):  # pragma: no cover
    """Output the data points of a data series to stdout."""
    # Print the "name" of the series:
    print('{} {} of {} in {}:'.format(data_series['frequency_name'],
                                      data_series['metric_name'],
                                      data_series['item_name'],
                                      data_series['region_name']))
    # Print the data points:
    for point in client.get_data_points(**data_series):
        print('{}-{}: {} {}'.format(point["start_date"],
                                    point["end_date"],
                                    point["value"],
                                    client.lookup_unit_abbreviation(point["unit_id"])))


def write_one_data_series(client, data_series, filename):  # pragma: no cover
    """Output the data points of a data series to a CSV file."""
    client.get_logger().warning("Using data series: {}".format(str(data_series)))
    client.get_logger().warning("Outputing to file: {}".format(filename))
    writer = unicodecsv.writer(open(filename, "wb"))
    for point in client.get_data_points(**data_series):
        writer.writerow(
            [
                point["start_date"],
                point["end_date"],
                point["value"],
                client.lookup_unit_abbreviation(point["unit_id"]),
            ]
        )


def main():  # pragma: no cover
    """Basic Gro API command line interface.

    Note that results are chosen randomly from matching selections, and so results are not
    deterministic. This tool is useful for simple queries, but anything more complex should be done
    using the provided Python packages.

    Usage examples:
        gro_client --item=soybeans  --region=brazil --partner_region china --metric export
        gro_client --item=sesame --region=ethiopia
        gro_client --user_email=john.doe@example.com  --print_token
    For more information use --help
    """
    parser = argparse.ArgumentParser(description="Gro API command line interface")
    parser.add_argument("--user_email")
    parser.add_argument("--user_password")
    parser.add_argument("--item")
    parser.add_argument("--metric")
    parser.add_argument("--region")
    parser.add_argument("--partner_region")
    parser.add_argument("--file")
    parser.add_argument(
        "--print_token",
        action="store_true",
        help="Ouput API access token for the given user email and password. "
        "Save it in GROAPI_TOKEN environment variable.",
    )
    parser.add_argument(
        "--token",
        default=os.environ.get("GROAPI_TOKEN"),
        help="Defaults to GROAPI_TOKEN environment variable.",
    )
    parser.add_argument("--version", action="store_true")
    args = parser.parse_args()

    if args.version:
        print(groclient.lib.get_version_info().get('api-client-version'))
        return

    assert (
        args.user_email or args.token
    ), "Need --token, or --user_email, or $GROAPI_TOKEN"
    access_token = None

    if args.token:
        access_token = args.token
    else:
        if not args.user_password:
            args.user_password = getpass.getpass()
        access_token = groclient.lib.get_access_token(
            groclient.cfg.API_HOST, args.user_email, args.user_password
        )

    if args.print_token:
        print(access_token)
        return

    client = GroClient(groclient.cfg.API_HOST, access_token)

    if (
        not args.metric
        and not args.item
        and not args.region
        and not args.partner_region
    ):
        data_series = pick_random_data_series(client)
    else:
        data_series = next(
            client.find_data_series(
                item=args.item,
                metric=args.metric,
                region=args.region,
                partner_region=args.partner_region,
            ),
            None
        )

    if data_series is None:
        print("No data series found.")
        return

    if args.file is not None:
        write_one_data_series(client, data_series, args.file)
    else:
        print_one_data_series(client, data_series)



"""Constants about the Gro ontology that can be imported and re-used anywhere."""

REGION_LEVELS = {
    'world': 1,
    'continent': 2,
    'country': 3,
    'province': 4,  # Equivalent to state in the United States
    'district': 5,  # Equivalent to county in the United States
    'city': 6,
    'market': 7,
    'other': 8,
    'coordinate': 9
}

ENTITY_TYPES_PLURAL = ['metrics', 'items', 'regions', 'frequencies', 'sources', 'units']

DATA_SERIES_UNIQUE_TYPES_ID = [
    'metric_id',
    'item_id',
    'region_id',
    'partner_region_id',
    'frequency_id',
    'source_id'
]

ENTITY_KEY_TO_TYPE = {'item_id': 'items',
                      'metric_id': 'metrics',
                      'region_id': 'regions',
                      'partner_region_id': 'regions',
                      'source_id': 'sources',
                      'frequency_id': 'frequencies'}

DATA_POINTS_UNIQUE_COLS = DATA_SERIES_UNIQUE_TYPES_ID + [
    'reporting_date',
    'start_date',
    'end_date'
]



from __future__ import print_function
# Sample Gro API client that retrieves data series for Brazil soybeans
# yield from multiple sources.
#
# Usage example: If you have saved GROAPI_TOKEN environment variable,
#
#   python soybeans.py
#
# If you haven't saved the token, you can pass it via cmd line
#
#   python soybeans.py --token ...
#
# Ref https://app.gro-intelligence.com/#/displays/25894

import argparse
import os
import sys

import groclient.lib
from api.client.crop_model import CropModel


def add_brazil_soybeans_yield(model):
    """Adds data series for national level soybeans yield for
    Brazil. There are multiple series from different sources, we add them all."""
    entities = {}
    # Search for item and metric by name, and use the top result
    entities['item_id'] =  model.search_for_entity('items', "soybeans")
    entities['metric_id'] = model.search_for_entity('metrics', "yield mass/area")
    # There are many regions with "brazil" in the name, level = 3 is
    # the country
    for region in model.search_and_lookup('regions', "brazil"):
        if region['level'] == 3:
            entities['region_id'] = region['id']
    data_series_list = model.get_data_series(**entities)
    print("There are {} data series for {}.".format(len(data_series_list), entities))
    for data_series in data_series_list:
        # Look up the name of the source based on the source_id
        source_name = model.lookup('sources', data_series['source_id']).get('longName')
        print(u'{}: {} to {}'.format(source_name,
                                    data_series['start_date'], data_series['end_date']))
        model.add_single_data_series(data_series)
    return


def get_brazil_soybeans_weighted_ndvi(model):
    provinces = model.get_provinces('brazil')
    df = model.compute_crop_weighted_series(
        'soybeans', 'Production Quantity mass',
        'Vegetation NDVI', 'Vegetation Indices index',
        provinces)
    return df


def main():
    parser = argparse.ArgumentParser(description="Gro api client")
    parser.add_argument("--token", default=os.environ['GROAPI_TOKEN'])
    args = parser.parse_args()

    model = CropModel('api.gro-intelligence.com', os.environ['GROAPI_TOKEN'])
    add_brazil_soybeans_yield(model)
    cwdf = get_brazil_soybeans_weighted_ndvi(model)
    df = model.get_df()
    print("{} data frame with all data series as fetched".format(df.shape))
    print("Columns: {}".format(df.columns))
    print("{} data frame with crop weighted data series".format(cwdf.shape))
    print("Columns: {}".format(cwdf.columns))


if __name__ == "__main__":
    main()
